@@ -1,6 +1,7 @@
 #include "config.h"
 
 #include <gtk/gtk.h>
+#include <libgnome/libgnome.h>
 #include <gdk/gdkkeysyms.h>
 #include <string.h>
 #include <gdk/gdkx.h>
@@ -12,12 +13,50 @@
 
 static gboolean caps_lock_state = FALSE;
 
+static Display *
+get_parent_display (void)
+{
+  static gboolean tested = FALSE;
+  static Display *dsp = NULL;
+
+  if (tested)
+    return dsp;
+
+  tested = TRUE;
+
+  if (g_getenv ("GDM_PARENT_DISPLAY") != NULL)
+    {
+      char *old_xauth = g_strdup (g_getenv ("XAUTHORITY"));
+      if (g_getenv ("GDM_PARENT_XAUTHORITY") != NULL)
+        {
+	  gnome_setenv ("XAUTHORITY",
+			g_getenv ("GDM_PARENT_XAUTHORITY"), TRUE);
+	}
+      dsp = XOpenDisplay (g_getenv ("GDM_PARENT_DISPLAY"));
+      if (old_xauth != NULL)
+        gnome_setenv ("XAUTHORITY", old_xauth, TRUE);
+      else
+        gnome_unsetenv ("XAUTHORITY");
+      g_free (old_xauth);
+    }
+
+  return dsp;
+}
+
 static gboolean
 is_caps_lock_on (void)
 {
   unsigned int states;
-	
-  if (XkbGetIndicatorState (GDK_DISPLAY (), XkbUseCoreKbd, &states) != Success)
+  Display *dsp;
+
+  /* HACK! incredible hack, if this is set we get
+   * indicator state from the parent display, since we must be inside an
+   * Xnest */
+  dsp = get_parent_display ();
+  if (dsp == NULL)
+    dsp = GDK_DISPLAY ();
+
+  if (XkbGetIndicatorState (dsp, XkbUseCoreKbd, &states) != Success)
       return FALSE;
 
   return (states & ShiftMask) != 0;
@@ -70,8 +109,8 @@ greeter_item_capslock_setup (GtkWidget *window)
 {
   capslock_update (is_caps_lock_on ());
   
-  gtk_signal_connect (GTK_OBJECT (window), "key_press_event",
-		      GTK_SIGNAL_FUNC (cl_key_press_event), NULL);
+  g_signal_connect (G_OBJECT (window), "key_press_event",
+		    G_CALLBACK (cl_key_press_event), NULL);
   return TRUE;
 }
 
